@@ -25,12 +25,9 @@
 	speed reasons.
 
 	Input port emulation is done externally.
-
-  Bugs:
-	The 21MHz master clock penalty when executing(not explicitly branching) across a 8KiB page boundary is not emulated.
 */
 
-#include "mednafen/mednafen.h"
+#include <mednafen/mednafen.h>
 #include "huc6280.h"
 
 #include <string.h>
@@ -39,8 +36,6 @@
  #include        <trio/trio.h>
 #endif
 
-//#define IncPC()	{ PC++; if(!(PC & 0x1FFF)) printf("Crossing: %04x %02x\n", PC - 1, lastop); }
-
 #define LASTCYCLE /*assert(((P & I_FLAG) ? 0 : (uint32)~0) == PIMaskCache);*/ IRQSample = (IRQlow & IRQMask) & PIMaskCache; IFlagSample = P & I_FLAG; ADDCYC(1);
 
 void HuC6280::StealCycle(void)
@@ -195,7 +190,7 @@ INLINE void HuC6280::BBSi(const uint8 val, const unsigned int bitto)
 	      }	\
 	     } TPOSTFIX;
 
-#define SBC  if(P & T_FLAG) {puts("SET misuse"); } if(P & D_FLAG)	\
+#define SBC  if(P & D_FLAG)	\
 	     {		\
 	      uint32 c = (P & 1) ^ 1;	\
 	      uint32 l = A - x - c;	\
@@ -662,7 +657,7 @@ void HuC6280::IRQStatusWrite(unsigned int address, uint8 V)
  }
 }
 
-int HuC6280::StateAction(StateMem *sm, int load, int data_only)
+void HuC6280::StateAction(StateMem *sm, const unsigned load, const bool data_only)
 {
  uint16 tmp_PC = PC;
 
@@ -705,7 +700,7 @@ int HuC6280::StateAction(StateMem *sm, int load, int data_only)
   SFEND
  };
 
- int ret = MDFNSS_StateAction(sm, load, data_only, StateRegs, "CPU");
+ MDFNSS_StateAction(sm, load, data_only, StateRegs, "CPU");
 
  if(load)
  {
@@ -716,17 +711,6 @@ int HuC6280::StateAction(StateMem *sm, int load, int data_only)
   REDOSPEEDCACHE();
   REDOPIMCACHE();
  }
- return(ret);
-}
-
-void HuC6280::DumpMem(char *filename, uint32 start, uint32 end)
-{
- FILE *fp=fopen(filename, "wb");
-
- for(uint32 x = start; x <= end; x++)
-  fputc(RdMem(x), fp);
-
- fclose(fp);
 }
 
 void HuC6280::SetRegister(const unsigned int id, uint32 value)
