@@ -15,6 +15,13 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+/*
+  Games to test after changing code affecting CD reading and buffering:
+		Bedlam
+		Rise 2
+
+*/
+
 // TODO: async command counter and async command phase?
 /*
 
@@ -64,6 +71,24 @@ PS_CDC::~PS_CDC()
 
 }
 
+void PS_CDC::DMForceStop(void)
+{
+ PSRCounter = 0;
+
+ if((DriveStatus != DS_PAUSED && DriveStatus != DS_STOPPED) || PendingCommandPhase >= 2)
+ {
+  PendingCommand = 0x00;
+  PendingCommandCounter = 0;
+  PendingCommandPhase = 0;
+ }
+
+ HeaderBufValid = false;
+ DriveStatus = DS_STOPPED;
+ ClearAIP();
+ SectorPipe_Pos = SectorPipe_In = 0;
+ SectorsRead = 0;
+}
+
 void PS_CDC::SetDisc(bool tray_open, CDIF *cdif, const char *disc_id)
 {
  if(tray_open)
@@ -75,19 +100,7 @@ void PS_CDC::SetDisc(bool tray_open, CDIF *cdif, const char *disc_id)
 
  if(!Cur_CDIF)
  {
-  PSRCounter = 0;
-
-  if((DriveStatus != DS_PAUSED && DriveStatus != DS_STOPPED) || PendingCommandPhase >= 2)
-  {
-   PendingCommand = 0x00;
-   PendingCommandCounter = 0;
-   PendingCommandPhase = 0;
-  }
-
-  HeaderBufValid = false;
-  DriveStatus = DS_STOPPED;
-  ClearAIP();
-  SectorPipe_Pos = SectorPipe_In = 0;
+  DMForceStop();
  }
  else
  {
@@ -115,7 +128,7 @@ int32 PS_CDC::CalcNextEvent(void)
  if(PendingCommandCounter > 0 && next_event > PendingCommandCounter)
   next_event = PendingCommandCounter;
 
- if(!IRQBuffer)
+ if(!(IRQBuffer & 0xF))
  {
   if(CDCReadyReceiveCounter > 0 && next_event > CDCReadyReceiveCounter)
    next_event = CDCReadyReceiveCounter;
@@ -158,6 +171,7 @@ void PS_CDC::SoftReset(void)
  DMABuffer.Flush();
  SB_In = 0;
  SectorPipe_Pos = SectorPipe_In = 0;
+ SectorsRead = 0;
 
  memset(SubQBuf, 0, sizeof(SubQBuf));
  memset(SubQBuf_Safe, 0, sizeof(SubQBuf_Safe));
@@ -173,7 +187,7 @@ void PS_CDC::SoftReset(void)
  PendingCommandPhase = 0;
  PendingCommandCounter = 0;
 
- Mode = 0;
+ Mode = 0x20;
 
  HeaderBufValid = false;
  DriveStatus = DS_STOPPED;
@@ -212,7 +226,7 @@ void PS_CDC::Power(void)
  lastts = 0;
 }
 
-int PS_CDC::StateAction(StateMem *sm, int load, int data_only)
+void PS_CDC::StateAction(StateMem *sm, const unsigned load, const bool data_only)
 {
  SFORMAT StateRegs[] =
  {
@@ -299,6 +313,7 @@ int PS_CDC::StateAction(StateMem *sm, int load, int data_only)
   SFVAR(PSRCounter),
 
   SFVAR(CurSector),
+  SFVAR(SectorsRead),
 
   SFVAR(AsyncIRQPending),
   SFARRAY(AsyncResultsPending, sizeof(AsyncResultsPending) / sizeof(AsyncResultsPending[0])),
@@ -326,14 +341,18 @@ int PS_CDC::StateAction(StateMem *sm, int load, int data_only)
   SFEND
  };
 
- int ret = MDFNSS_StateAction(sm, load, data_only, StateRegs, "CDC");
+ MDFNSS_StateAction(sm, load, data_only, StateRegs, "CDC");
 
  if(load)
  {
   DMABuffer.SaveStatePostLoad();
   SectorPipe_Pos %= SectorPipe_Count;
+
+  //
+  // Handle pre-0.9.37 state loading, and maliciously-constructed/corrupted save states.
+  if(!Cur_CDIF)
+   DMForceStop();
  }
- return(ret);
 }
 
 void PS_CDC::ResetTS(void)
@@ -345,17 +364,18 @@ void PS_CDC::RecalcIRQ(void)
 {
  IRQ_Assert(IRQ_CD, (bool)(IRQBuffer & (IRQOutTestMask & 0x1F)));
 }
+
 //static int32 doom_ts;
 void PS_CDC::WriteIRQ(uint8 V)
 {
  assert(CDCReadyReceiveCounter <= 0);
- assert(!IRQBuffer);
+ assert(!(IRQBuffer & 0xF));
 
  //PSX_WARNING("[CDC] ***IRQTHINGY: 0x%02x -- %u", V, doom_ts);
 
  CDCReadyReceiveCounter = 2000; //1024;
 
- IRQBuffer = V;
+ IRQBuffer = (IRQBuffer & 0x10) | V;
  RecalcIRQ();
 }
 
@@ -366,10 +386,11 @@ void PS_CDC::BeginResults(void)
  // printf("Cleared %d results. IRQBuffer=0x%02x\n", ResultsIn, IRQBuffer);
  //}
 
- // TODO: test semantics on real thing.
  ResultsIn = 0;
  ResultsWP = 0;
  ResultsRP = 0;
+
+ memset(ResultsBuffer, 0x00, sizeof(ResultsBuffer));
 }
 
 void PS_CDC::WriteResult(uint8 V)
@@ -386,6 +407,9 @@ uint8 PS_CDC::ReadResult(void)
 {
  uint8 ret = ResultsBuffer[ResultsRP];
 
+ if(!ResultsIn)
+  PSX_WARNING("[CDC] Results buffer underflow!");
+
  ResultsRP = (ResultsRP + 1) & 0xF;
  ResultsIn = (ResultsIn - 1) & 0x1F;
 
@@ -401,11 +425,17 @@ uint8 PS_CDC::MakeStatus(bool cmd_error)
  if(DriveStatus == DS_PLAYING)
   ret |= 0x80;
 
- if(DriveStatus == DS_SEEKING || DriveStatus == DS_SEEKING_LOGICAL)
-  ret |= 0x40;
-
+ // Probably will want to be careful with this HeaderBufValid versus seek/read bit business in the future as it is a bit fragile;
+ // "Gran Turismo 1"'s music(or erroneous lack of) is a good test case.
  if(DriveStatus == DS_READING)
-  ret |= 0x20;
+ {
+  if(!HeaderBufValid)
+   ret |= 0x40;
+  else
+   ret |= 0x20;
+ }
+ else if(DriveStatus == DS_SEEKING || DriveStatus == DS_SEEKING_LOGICAL)
+  ret |= 0x40;
 
  // TODO: shell open and seek error
  if(!Cur_CDIF || DiscChanged)
@@ -857,25 +887,17 @@ void PS_CDC::HandlePlayRead(void)
   PSX_WARNING("[CDC] Read/Play position waaay too far out(%u), forcing STOP", CurSector);
   DriveStatus = DS_STOPPED;
   SectorPipe_Pos = SectorPipe_In = 0;
+  SectorsRead = 0;
   return;
  }
 
  if(CurSector >= (int32)toc.tracks[100].lba)
  {
   PSX_WARNING("[CDC] In leadout area: %u", CurSector);
+ }
 
-  //
-  // Synthesis is a bit of a kludge... :/
-  //
-  synth_leadout_sector_lba(0x02, toc, CurSector, read_buf);
-  DecodeSubQ(read_buf + 2352);
- }
- else
- {
-  Cur_CDIF->ReadRawSector(read_buf, CurSector);	// FIXME: error out on error.
-  DecodeSubQ(read_buf + 2352);
- }
- 
+ Cur_CDIF->ReadRawSector(read_buf, CurSector);	// FIXME: error out on error.
+ DecodeSubQ(read_buf + 2352);
 
  if(SubQBuf_Safe[1] == 0xAA && (DriveStatus == DS_PLAYING || (!(SubQBuf_Safe[0] & 0x40) && (Mode & MODE_CDDA))))
  {
@@ -886,6 +908,7 @@ void PS_CDC::HandlePlayRead(void)
   // Status in this end-of-disc context here should be generated after we're in the pause state.
   DriveStatus = DS_PAUSED;
   SectorPipe_Pos = SectorPipe_In = 0;
+  SectorsRead = 0;
   SetAIP(CDCIRQ_DATA_END, MakeStatus());
 
   return;
@@ -904,6 +927,7 @@ void PS_CDC::HandlePlayRead(void)
 
    DriveStatus = DS_PAUSED;
    SectorPipe_Pos = SectorPipe_In = 0;
+   SectorsRead = 0;
    PSRCounter = 0;
    return;
   }
@@ -1039,6 +1063,7 @@ void PS_CDC::HandlePlayRead(void)
  else
   CurSector++;
 
+ SectorsRead++;
 }
 
 pscpu_timestamp_t PS_CDC::Update(const pscpu_timestamp_t timestamp)
@@ -1075,7 +1100,7 @@ pscpu_timestamp_t PS_CDC::Update(const pscpu_timestamp_t timestamp)
 
   //MDFN_DispMessage("%02x %d -- %d %d -- %02x", IRQBuffer, CDCReadyReceiveCounter, PSRCounter, PendingCommandCounter, PendingCommand);
 
-  if(!IRQBuffer)
+  if(!(IRQBuffer & 0xF))
   {
    if(CDCReadyReceiveCounter > 0 && chunk_clocks > CDCReadyReceiveCounter)
     chunk_clocks = CDCReadyReceiveCounter;
@@ -1088,7 +1113,7 @@ pscpu_timestamp_t PS_CDC::Update(const pscpu_timestamp_t timestamp)
 
   if(PSRCounter > 0)
   {
-   uint8 buf[2352 + 96];
+   uint8 pwbuf[96];
 
    PSRCounter -= chunk_clocks;
 
@@ -1103,8 +1128,9 @@ pscpu_timestamp_t PS_CDC::Update(const pscpu_timestamp_t timestamp)
 
      SB_In = 0;
      SectorPipe_Pos = SectorPipe_In = 0;
+     SectorsRead = 0;
 
-     Mode = 0;
+     Mode = 0x20;	// Confirmed(and see "This Is Football 2").
      CurSector = 0;
      CommandLoc = 0;
 
@@ -1114,8 +1140,18 @@ pscpu_timestamp_t PS_CDC::Update(const pscpu_timestamp_t timestamp)
     else if(DriveStatus == DS_SEEKING)
     {
      CurSector = SeekTarget;
-     Cur_CDIF->ReadRawSector(buf, CurSector);
-     DecodeSubQ(buf + 2352);
+
+     //
+     // CurSector + x for "Tomb Raider"'s sake, as it relies on behavior that we can't emulate very well without a more accurate CD drive
+     // emulation model.
+     //
+     for(int x = -1; x >= -16; x--)
+     {
+      //printf("%d\n", CurSector + x);
+      Cur_CDIF->ReadRawSectorPWOnly(pwbuf, CurSector + x, false);
+      if(DecodeSubQ(pwbuf))
+       break;
+     }
 
      DriveStatus = StatusAfterSeek;
 
@@ -1127,9 +1163,8 @@ pscpu_timestamp_t PS_CDC::Update(const pscpu_timestamp_t timestamp)
     else if(DriveStatus == DS_SEEKING_LOGICAL)
     {
      CurSector = SeekTarget;
-     Cur_CDIF->ReadRawSector(buf, CurSector);
-     DecodeSubQ(buf + 2352);
-     memcpy(HeaderBuf, buf + 12, 12);
+     Cur_CDIF->ReadRawSectorPWOnly(pwbuf, CurSector, false);
+     DecodeSubQ(pwbuf);
 
      DriveStatus = StatusAfterSeek;
 
@@ -1291,7 +1326,7 @@ void PS_CDC::Write(const pscpu_timestamp_t timestamp, uint32 A, uint8 V)
 			PendingCommandCounter, V);
 		}
 
-		if(IRQBuffer)
+		if(IRQBuffer & 0xF)
 		{
 		 PSX_WARNING("[CDC] Attempting to start command(0x%02x) while IRQBuffer(0x%02x) is not clear.", V, IRQBuffer);
 		}
@@ -1356,6 +1391,7 @@ void PS_CDC::Write(const pscpu_timestamp_t timestamp, uint32 A, uint8 V)
 		{
 		 PSX_WARNING("[CDC] Mystery IRQ trigger bit set.");
 		 IRQBuffer |= 0x10;
+		 RecalcIRQ();
 		}
 		break;
 
@@ -1500,7 +1536,7 @@ uint32 PS_CDC::DMARead(void)
    data |= DMABuffer.ReadByte() << (i * 8);
   else
   {
-   //assert(0);
+   PSX_WARNING("[CDC] DMA read buffer underflow!");
   }
  }
 
@@ -1522,12 +1558,6 @@ bool PS_CDC::CommandCheckDiscPresent(void)
  return(true);
 }
 
-int32 PS_CDC::Command_Sync(const int arg_count, const uint8 *args)
-{
- PSX_WARNING("[CDC] Unimplemented command: 0x%02x", PendingCommand);
- return(0);
-}
-
 int32 PS_CDC::Command_Nop(const int arg_count, const uint8 *args)
 {
  WriteResult(MakeStatus());
@@ -1540,7 +1570,17 @@ int32 PS_CDC::Command_Setloc(const int arg_count, const uint8 *args)
 {
  uint8 m, s, f;
 
- m = BCD_to_U8(args[0] & 0x7F);
+ if((args[0] & 0x0F) > 0x09 || args[0] > 0x99 ||
+    (args[1] & 0x0F) > 0x09 || args[1] > 0x59 ||
+    (args[2] & 0x0F) > 0x09 || args[2] > 0x74)
+ {
+  WriteResult(MakeStatus(true));
+  WriteResult(ERRCODE_BAD_ARGVAL);
+  WriteIRQ(CDCIRQ_DISC_ERROR);
+  return(0);
+ }
+
+ m = BCD_to_U8(args[0]);
  s = BCD_to_U8(args[1]);
  f = BCD_to_U8(args[2]);
 
@@ -1587,10 +1627,12 @@ int32 PS_CDC::CalcSeekTime(int32 initial, int32 target, bool motor_on, bool paus
    ret += 1237952 * ((Mode & MODE_SPEED) ? 1 : 2);
   }
  }
+ //else if(target < initial)
+ // ret += 1000000;
 
  ret += PSX_GetRandU32(0, 25000);
 
- PSX_DBG(PSX_DBG_SPARSE, "[CDC] CalcSeekTime() = %d\n", ret);
+ PSX_DBG(PSX_DBG_SPARSE, "[CDC] CalcSeekTime() %d->%d = %d\n", initial, target, ret);
 
  return(ret);
 }
@@ -1607,32 +1649,24 @@ void PS_CDC::BeginSeek(uint32 target, int after_seek)
 
 // Remove this function when we have better seek emulation; it's here because the Rockman complete works games(at least 2 and 4) apparently have finicky fubared CD
 // access code.
-void PS_CDC::PreSeekHack(bool logical, uint32 target)
+void PS_CDC::PreSeekHack(int32 target)
 {
- uint8 buf[2352 + 96];
+ uint8 pwbuf[96];
  int max_try = 32;
- bool NeedHBuf = logical;
 
  CurSector = target;	// If removing/changing this, take into account how it will affect ReadN/ReadS/Play/etc command calls that interrupt a seek.
 
  // If removing this SubQ reading bit, think about how it will interact with a Read command of data(or audio :b) sectors when Mode bit0 is 1.
- if(target < toc.tracks[100].lba)
+ do
  {
-  do
-  {
-   Cur_CDIF->ReadRawSector(buf, target++);
-
-   // GetLocL related kludge, for Gran Turismo 1 music, perhaps others?
-   if(NeedHBuf)
-   {
-    NeedHBuf = false;
-    memcpy(HeaderBuf, buf + 12, 12);
-    HeaderBufValid = true;
-   }
-  } while(!DecodeSubQ(buf + 2352) && --max_try > 0 && target < toc.tracks[100].lba);
- }
+  Cur_CDIF->ReadRawSectorPWOnly(pwbuf, target++, true);
+ } while(!DecodeSubQ(pwbuf) && --max_try > 0);
 }
 
+/*
+ Play command with a track argument that's not a valid BCD quantity causes interesting half-buggy behavior on an actual PS1(unlike some of the other commands,
+ an error doesn't seem to be generated for a bad BCD argument).
+*/
 int32 PS_CDC::Command_Play(const int arg_count, const uint8 *args)
 {
  if(!CommandCheckDiscPresent())
@@ -1656,12 +1690,13 @@ int32 PS_CDC::Command_Play(const int arg_count, const uint8 *args)
   }
   else if(track > toc.last_track)
   {
-   PSX_WARNING("[CDC] Attempt to play track before first track.");
+   PSX_WARNING("[CDC] Attempt to play track after last track.");
    track = toc.last_track;
   }
 
   ClearAudioBuffers();
   SectorPipe_Pos = SectorPipe_In = 0;
+  SectorsRead = 0;
 
   PlayTrackMatch = track;
 
@@ -1670,7 +1705,7 @@ int32 PS_CDC::Command_Play(const int arg_count, const uint8 *args)
   SeekTarget = toc.tracks[track].lba;
   PSRCounter = CalcSeekTime(CurSector, SeekTarget, DriveStatus != DS_STOPPED, DriveStatus == DS_PAUSED);
   HeaderBufValid = false;
-  PreSeekHack(false, SeekTarget);
+  PreSeekHack(SeekTarget);
 
   ReportLastF = 0xFF;
 
@@ -1681,6 +1716,7 @@ int32 PS_CDC::Command_Play(const int arg_count, const uint8 *args)
  {
   ClearAudioBuffers();
   SectorPipe_Pos = SectorPipe_In = 0;
+  SectorsRead = 0;
 
   if(CommandLoc_Dirty)
    SeekTarget = CommandLoc;
@@ -1691,7 +1727,7 @@ int32 PS_CDC::Command_Play(const int arg_count, const uint8 *args)
 
   PSRCounter = CalcSeekTime(CurSector, SeekTarget, DriveStatus != DS_STOPPED, DriveStatus == DS_PAUSED);
   HeaderBufValid = false;
-  PreSeekHack(false, SeekTarget);
+  PreSeekHack(SeekTarget);
 
   ReportLastF = 0xFF;
 
@@ -1753,6 +1789,7 @@ void PS_CDC::ReadBase(void)
   ClearAudioBuffers();
   SB_In = 0;
   SectorPipe_Pos = SectorPipe_In = 0;
+  SectorsRead = 0;
 
   // TODO: separate motor start from seek phase?
 
@@ -1763,7 +1800,7 @@ void PS_CDC::ReadBase(void)
 
   PSRCounter = /*903168 * 1.5 +*/ CalcSeekTime(CurSector, SeekTarget, DriveStatus != DS_STOPPED, DriveStatus == DS_PAUSED);
   HeaderBufValid = false;
-  PreSeekHack(true, SeekTarget);
+  PreSeekHack(SeekTarget);
 
   DriveStatus = DS_SEEKING_LOGICAL;
   StatusAfterSeek = DS_READING;
@@ -1801,6 +1838,7 @@ int32 PS_CDC::Command_Stop(const int arg_count, const uint8 *args)
   ClearAudioBuffers();
   ClearAIP();
   SectorPipe_Pos = SectorPipe_In = 0;
+  SectorsRead = 0;
 
   DriveStatus = DS_STOPPED;
   HeaderBufValid = false;
@@ -1838,6 +1876,7 @@ int32 PS_CDC::Command_Standby(const int arg_count, const uint8 *args)
  ClearAudioBuffers();
  ClearAIP();
  SectorPipe_Pos = SectorPipe_In = 0;
+ SectorsRead = 0;
 
  DriveStatus = DS_STANDBY;
 
@@ -1868,6 +1907,9 @@ int32 PS_CDC::Command_Pause(const int arg_count, const uint8 *args)
  }
  else
  {
+  CurSector -= std::min<uint32>(4, SectorsRead);	// See: Bedlam, Rise 2
+  SectorsRead = 0;
+
   // "Viewpoint" flips out and crashes if reading isn't stopped (almost?) immediately.
   //ClearAudioBuffers();
   SectorPipe_Pos = SectorPipe_In = 0;
@@ -2044,16 +2086,17 @@ int32 PS_CDC::Command_GetTD(const int arg_count, const uint8 *args)
  int track;
  uint8 m, s, f;
 
- if(!args[0] || args[0] == 0xAA)
+ if(!args[0])
   track = 100;
  else
  {
-  track= BCD_to_U8(args[0]);
+  track = BCD_to_U8(args[0]);
 
-  if(track < toc.first_track || track > toc.last_track)	// Error
+  if(!BCD_is_valid(args[0]) || track < toc.first_track || track > toc.last_track)	// Error
   {
    WriteResult(MakeStatus(true));
-   WriteIRQ(CDCIRQ_ACKNOWLEDGE);
+   WriteResult(ERRCODE_BAD_ARGVAL);
+   WriteIRQ(CDCIRQ_DISC_ERROR);
    return(0);
   }
  }
@@ -2080,9 +2123,9 @@ int32 PS_CDC::Command_SeekL(const int arg_count, const uint8 *args)
 
  SeekTarget = CommandLoc;
 
- PSRCounter = CalcSeekTime(CurSector, SeekTarget, DriveStatus != DS_STOPPED, DriveStatus == DS_PAUSED);
+ PSRCounter = (33868800 / (75 * ((Mode & MODE_SPEED) ? 2 : 1))) + CalcSeekTime(CurSector, SeekTarget, DriveStatus != DS_STOPPED, DriveStatus == DS_PAUSED);
  HeaderBufValid = false;
- PreSeekHack(true, SeekTarget);
+ PreSeekHack(SeekTarget);
  DriveStatus = DS_SEEKING_LOGICAL;
  StatusAfterSeek = DS_STANDBY;
  ClearAIP();
@@ -2102,7 +2145,7 @@ int32 PS_CDC::Command_SeekP(const int arg_count, const uint8 *args)
 
  PSRCounter = CalcSeekTime(CurSector, SeekTarget, DriveStatus != DS_STOPPED, DriveStatus == DS_PAUSED);
  HeaderBufValid = false;
- PreSeekHack(false, SeekTarget);
+ PreSeekHack(SeekTarget);
  DriveStatus = DS_SEEKING;
  StatusAfterSeek = DS_STANDBY;
  ClearAIP();
@@ -2357,7 +2400,7 @@ int32 PS_CDC::Command_0x1d(const int arg_count, const uint8 *args)
 
 PS_CDC::CDC_CTEntry PS_CDC::Commands[0x20] =
 {
- { /* 0x00, */ 0, 0, "Sync", &PS_CDC::Command_Sync, NULL },
+ { /* 0x00, */ 0, 0, NULL, NULL, NULL },
  { /* 0x01, */ 0, 0, "Nop", &PS_CDC::Command_Nop, NULL },
  { /* 0x02, */ 3, 3, "Setloc", &PS_CDC::Command_Setloc, NULL },
  { /* 0x03, */ 0, 1, "Play", &PS_CDC::Command_Play, NULL },
