@@ -25,7 +25,7 @@ class PS_CDC
  void SetDisc(bool tray_open, CDIF *cdif, const char disc_id[4]);
 
  void Power(void);
- int StateAction(StateMem *sm, int load, int data_only);
+ void StateAction(StateMem *sm, const unsigned load, const bool data_only);
  void ResetTS(void);
 
  int32 CalcNextEvent(void);	// Returns in master cycles to next event.
@@ -171,6 +171,7 @@ class PS_CDC
  int32 PSRCounter;
 
  int32 CurSector;
+ uint32 SectorsRead;	// Reset to 0 on Read*/Play command start; used in the rough simulation of PS1 SetLoc->Read->Pause->Read behavior.
 
  unsigned AsyncIRQPending;
  uint8 AsyncResultsPending[16];
@@ -198,6 +199,7 @@ class PS_CDC
  uint8 MakeStatus(bool cmd_error = false);
  bool DecodeSubQ(uint8 *subpw);
  bool CommandCheckDiscPresent(void);
+ void DMForceStop();
 
  void EnbufferizeCDDASector(const uint8 *buf);
  bool XA_Test(const uint8 *sdata);
@@ -220,13 +222,11 @@ class PS_CDC
   int32 (PS_CDC::*func2)(void);
  };
 
- void BeginSeek(uint32 target);
- void PreSeekHack(bool logical, uint32 target);
+ void PreSeekHack(int32 target);
  void ReadBase(void);
 
  static CDC_CTEntry Commands[0x20];
 
- int32 Command_Sync(const int arg_count, const uint8 *args);
  int32 Command_Nop(const int arg_count, const uint8 *args);
  int32 Command_Setloc(const int arg_count, const uint8 *args);
  int32 Command_Play(const int arg_count, const uint8 *args);
