@@ -18,21 +18,6 @@
 #include "psx.h"
 #include "timer.h"
 
-/*
- TODO:
-	Test and clean up line, particularly polyline, drawing.
-
-	"abe" transparency testing might be not correct, the transparency in regards to mask bit setting and evaluation may not be correct.
-
-	Not everything is returned in the status port read yet(double check).
-
-	"dfe" bit of drawing mode is probably not implemented 100% correctly.
-
-	Initialize more stuff in the Power() function.
-
-	Fix triangle span rendering order(it's bottom-to-up sometimes on the real thing, to avoid negative x step/increment values).
-*/
-
 /*
  GPU display timing master clock is nominally 53.693182 MHz for NTSC PlayStations, and 53.203425 MHz for PAL PlayStations.
 
@@ -58,24 +43,6 @@
 
 */
 
-/*
- Known problematic games to do regression testing on:
-
-	Dukes of Hazzard: Racing For Home
-		Sensitive about the GPU busy status flag being set long enough; double-check if we ever make CPU emulation more timing-accurate(
-		the fix will likely just involve reducing the timing granularity for DMA and GPU updates).
-
-	Final Fantasy 7
-		WHERE DO I BEGIN?!  (Currently broken as of Jan. 1, 2012)
-
-	Pro Pinball (series)
-		Sensitive to correct interlace and draw line skipping emulation.
-
-	Valkyrie Profile
-		Battle scenes will go all kaka with no graphics updates if GPU LL DMA completes too soon.
-
-*/
-
 /*
  November 29, 2012 notes:
 
@@ -90,9 +57,8 @@
 
 namespace MDFN_IEN_PSX
 {
-//FILE *fp;
 
-static const int32 dither_table[4][4] =
+static const int8 dither_table[4][4] =
 {
  { -4,  0, -3,  1 },
  {  2, -2,  3, -1 },
@@ -100,10 +66,12 @@ static const int32 dither_table[4][4] =
  {  3, -1,  2, -2 },
 };
 
-PS_GPU::PS_GPU(bool pal_clock_and_tv, int sls, int sle) : BlitterFIFO(0x20) // 0x10 on actual PS1 GPU, 0x20 here(see comment at top of gpu.h)	// 0x10)
+PS_GPU::PS_GPU(bool pal_clock_and_tv, int sls, int sle, bool show_h_overscan)
 {
  HardwarePALType = pal_clock_and_tv;
 
+ hide_hoverscan = !show_h_overscan;
+
  for(int y = 0; y < 4; y++)
   for(int x = 0; x < 4; x++)
    for(int v = 0; v < 512; v++)
@@ -124,21 +92,23 @@ PS_GPU::PS_GPU(bool pal_clock_and_tv, int sls, int sle) : BlitterFIFO(0x20) // 0
  if(HardwarePALType == false)	// NTSC clock
  {
   GPUClockRatio = 103896; // 65536 * 53693181.818 / (44100 * 768)
+  hmc_to_visible = 520; 
  }
  else	// PAL clock
  {
   GPUClockRatio = 102948; // 65536 * 53203425 / (44100 * 768)
+  hmc_to_visible = 560; 
  }
 
- memset(RGB8SAT_Under, 0, sizeof(RGB8SAT_Under));
-
- for(int i = 0; i < 256; i++)
-  RGB8SAT[i] = i;
-
- memset(RGB8SAT_Over, 0xFF, sizeof(RGB8SAT_Over));
-
  LineVisFirst = sls;
  LineVisLast = sle;
+
+
+ memcpy(&Commands[0x00], Commands_00_1F, sizeof(Commands_00_1F));
+ memcpy(&Commands[0x20], Commands_20_3F, sizeof(Commands_20_3F));
+ memcpy(&Commands[0x40], Commands_40_5F, sizeof(Commands_40_5F));
+ memcpy(&Commands[0x60], Commands_60_7F, sizeof(Commands_60_7F));
+ memcpy(&Commands[0x80], Commands_80_FF, sizeof(Commands_80_FF));
 }
 
 PS_GPU::~PS_GPU()
@@ -150,10 +120,10 @@ void PS_GPU::FillVideoParams(MDFNGI* gi)
 {
  if(HardwarePALType)
  {
-  gi->lcm_width = 2800;
+  gi->lcm_width = hide_hoverscan ? 2640 : 2800;
   gi->lcm_height = (LineVisLast + 1 - LineVisFirst) * 2; //576;
 
-  gi->nominal_width = 377;	// Dunno. :(
+  gi->nominal_width = hide_hoverscan ? 363 : 384;	// More like 385.stuff according to calculations derived from BT.601, but 384 is a nicer number. :p
   gi->nominal_height = LineVisLast + 1 - LineVisFirst; //288;
 
   gi->fb_width = 768;
@@ -165,10 +135,10 @@ void PS_GPU::FillVideoParams(MDFNGI* gi)
  }
  else
  {
-  gi->lcm_width = 2800;
+  gi->lcm_width = hide_hoverscan ? 2640 : 2800;
   gi->lcm_height = (LineVisLast + 1 - LineVisFirst) * 2; //480;
 
-  gi->nominal_width = 320;
+  gi->nominal_width = (hide_hoverscan ? 302 : 320);
   gi->nominal_height = LineVisLast + 1 - LineVisFirst; //240;
 
   gi->fb_width = 768;
@@ -184,7 +154,7 @@ void PS_GPU::FillVideoParams(MDFNGI* gi)
  // For Justifier and Guncon.
  //
  gi->mouse_scale_x = (float)gi->lcm_width / gi->nominal_width;
- gi->mouse_offs_x = 0;
+ gi->mouse_offs_x = (float)(2800 - gi->lcm_width) / 2;
 
  gi->mouse_scale_y = 1.0;
  gi->mouse_offs_y = LineVisFirst;
@@ -195,39 +165,28 @@ void PS_GPU::SoftReset(void) // Control command 0x00
  IRQPending = false;
  IRQ_Assert(IRQ_GPU, IRQPending);
 
+ InvalidateCache();
+
  DMAControl = 0;
 
  if(DrawTimeAvail < 0)
   DrawTimeAvail = 0;
 
  BlitterFIFO.Flush();
+ DataReadBufferEx = 0;
  InCmd = INCMD_NONE;
 
  DisplayOff = 1;
  DisplayFB_XStart = 0;
  DisplayFB_YStart = 0;
 
- if(HardwarePALType)
- {
-  DisplayMode = 0x08;
+ DisplayMode = 0;
 
-  // FIXME, timing values(I need a PAL PS1 to derive them); for now, just copy the NTSC ones.
-  HorizStart = 0x200;
-  HorizEnd = 0xC00;
+ HorizStart = 0x200;
+ HorizEnd = 0xC00;
 
-  VertStart = 0x10;
-  VertEnd = 0x100;
- }
- else
- {
-  DisplayMode = 0;
-
-  HorizStart = 0x200;
-  HorizEnd = 0xC00;
-
-  VertStart = 0x10;
-  VertEnd = 0x100;
- }
+ VertStart = 0x10;
+ VertEnd = 0x100;
 
  //
  TexPageX = 0;
@@ -247,7 +206,7 @@ void PS_GPU::SoftReset(void) // Control command 0x00
  twx = 0;
  twy = 0;
 
- RecalcTexWindowLUT();
+ RecalcTexWindowStuff();
 
  //
  ClipX0 = 0;
@@ -264,18 +223,26 @@ void PS_GPU::SoftReset(void) // Control command 0x00
  //
  MaskSetOR = 0;
  MaskEvalAND = 0;
+
+ TexDisable = false;
+ TexDisableAllowChange = false;
 }
 
 void PS_GPU::Power(void)
 {
  memset(GPURAM, 0, sizeof(GPURAM));
 
+ memset(CLUT_Cache, 0, sizeof(CLUT_Cache));
+ CLUT_Cache_VB = ~0U;
+
+ memset(TexCache, 0xFF, sizeof(TexCache));
+
  DMAControl = 0;
 
  ClipX0 = 0;
  ClipY0 = 0;
- ClipX1 = 1023;
- ClipY1 = 1023;
+ ClipX1 = 0;
+ ClipY1 = 0;
 
  OffsX = 0;
  OffsY = 0;
@@ -286,13 +253,14 @@ void PS_GPU::Power(void)
  MaskSetOR = 0;
  MaskEvalAND = 0;
 
+ TexDisable = false;
+ TexDisableAllowChange = false;
+
  tww = 0;
  twh = 0;
  twx = 0;
  twy = 0;
 
- RecalcTexWindowLUT();
-
  TexPageX = 0;
  TexPageY = 0;
  SpriteFlip = 0;
@@ -300,8 +268,11 @@ void PS_GPU::Power(void)
  abr = 0;
  TexMode = 0;
 
- BlitterFIFO.Flush();
+ RecalcTexWindowStuff();
 
+ BlitterFIFO.Flush();
+ DataReadBuffer = 0;	// Don't reset in SoftReset()
+ DataReadBufferEx = 0;
  InCmd = INCMD_NONE;
  FBRW_X = 0;
  FBRW_Y = 0;
@@ -362,129 +333,7 @@ void PS_GPU::ResetTS(void)
  lastts = 0;
 }
 
-template<int BlendMode, bool MaskEval_TA, bool textured>
-INLINE void PS_GPU::PlotPixel(int32 x, int32 y, uint16 fore_pix)
-{
- y &= 511;	// More Y precision bits than GPU RAM installed in (non-arcade, at least) Playstation hardware.
-
- if(BlendMode >= 0 && (fore_pix & 0x8000))
- {
-  uint16 bg_pix = GPURAM[y][x];	// Don't use bg_pix for mask evaluation, it's modified in blending code paths.
-  uint16 pix; // = fore_pix & 0x8000;
-
-/*
- static const int32 tab[4][2] =
- {
-  { 2,  2 },
-  { 4,  4 },
-  { 4, -4 },
-  { 4,  1 }
- };
-*/
-  // Efficient 15bpp pixel math algorithms from blargg
-  switch(BlendMode)
-  {
-   case 0:
-	bg_pix |= 0x8000;
-	pix = ((fore_pix + bg_pix) - ((fore_pix ^ bg_pix) & 0x0421)) >> 1;
-	break;
-	  
-   case 1:
-       {
-	bg_pix &= ~0x8000;
-
-	uint32 sum = fore_pix + bg_pix;
-	uint32 carry = (sum - ((fore_pix ^ bg_pix) & 0x8421)) & 0x8420;
-
-	pix = (sum - carry) | (carry - (carry >> 5));
-       }
-       break;
-
-   case 2:
-       {
-	bg_pix |= 0x8000;
-        fore_pix &= ~0x8000;
-
-	uint32 diff = bg_pix - fore_pix + 0x108420;
-	uint32 borrow = (diff - ((bg_pix ^ fore_pix) & 0x108420)) & 0x108420;
-
-	pix = (diff - borrow) & (borrow - (borrow >> 5));
-       }
-       break;
-
-   case 3:
-       {
-	bg_pix &= ~0x8000;
-	fore_pix = ((fore_pix >> 2) & 0x1CE7) | 0x8000;
-
-	uint32 sum = fore_pix + bg_pix;
-	uint32 carry = (sum - ((fore_pix ^ bg_pix) & 0x8421)) & 0x8420;
-
-	pix = (sum - carry) | (carry - (carry >> 5));
-       }
-       break;
-  }
-
-  if(!MaskEval_TA || !(GPURAM[y][x] & 0x8000))
-   GPURAM[y][x] = (textured ? pix : (pix & 0x7FFF)) | MaskSetOR;
- }
- else
- {
-  if(!MaskEval_TA || !(GPURAM[y][x] & 0x8000))
-   GPURAM[y][x] = (textured ? fore_pix : (fore_pix & 0x7FFF)) | MaskSetOR;
- }
-}
-
-INLINE uint16 PS_GPU::ModTexel(uint16 texel, int32 r, int32 g, int32 b, const int32 dither_x, const int32 dither_y)
-{
- uint16 ret = texel & 0x8000;
-
- ret |= DitherLUT[dither_y][dither_x][(((texel & 0x1F) * r) >> (5 - 1))] << 0;
- ret |= DitherLUT[dither_y][dither_x][(((texel & 0x3E0) * g) >> (10 - 1))] << 5;
- ret |= DitherLUT[dither_y][dither_x][(((texel & 0x7C00) * b) >> (15 - 1))] << 10;
-
- return(ret);
-}
-
-template<uint32 TexMode_TA>
-INLINE uint16 PS_GPU::GetTexel(const uint32 clut_offset, int32 u_arg, int32 v_arg)
-{
-     uint32 u = TexWindowXLUT[u_arg];
-     uint32 v = TexWindowYLUT[v_arg];
-     uint32 fbtex_x = TexPageX + (u >> (2 - TexMode_TA));
-     uint32 fbtex_y = TexPageY + v;
-     uint16 fbw = GPURAM[fbtex_y][fbtex_x & 1023];
-
-     if(TexMode_TA != 2)
-     {
-      if(TexMode_TA == 0)
-       fbw = (fbw >> ((u & 3) * 4)) & 0xF;
-      else
-       fbw = (fbw >> ((u & 1) * 8)) & 0xFF;
- 
-      fbw = GPURAM[(clut_offset >> 10) & 511][(clut_offset + fbw) & 1023];
-     }
-
-     return(fbw);
-}
-
-INLINE bool PS_GPU::LineSkipTest(unsigned y)
-{
- //DisplayFB_XStart >= OffsX && DisplayFB_YStart >= OffsY &&
- // ((y & 1) == (DisplayFB_CurLineYReadout & 1))
-
- if((DisplayMode & 0x24) != 0x24)
-  return false;
-
- if(!dfe && ((y & 1) == ((DisplayFB_YStart + field_ram_readout) & 1))/* && !DisplayOff*/) //&& (y >> 1) >= DisplayFB_YStart && (y >> 1) < (DisplayFB_YStart + (VertEnd - VertStart)))
-  return true;
-
- return false;
-}
-
-#include "gpu_polygon.inc"
-#include "gpu_sprite.inc"
-#include "gpu_line.inc"
+#include "gpu_common.inc"
 
 // Special RAM write mode(16 pixels at a time), does *not* appear to use mask drawing environment settings.
 INLINE void PS_GPU::Command_FBFill(const uint32 *cb)
@@ -537,6 +386,7 @@ INLINE void PS_GPU::Command_FBCopy(const uint32 *cb)
  if(!height)
   height = 0x200;
 
+ InvalidateTexCache();
  //printf("FB Copy: %d %d %d %d %d %d\n", sourceX, sourceY, destX, destY, width, height);
 
  DrawTimeAvail -= (width * height) * 2;
@@ -546,7 +396,7 @@ INLINE void PS_GPU::Command_FBCopy(const uint32 *cb)
   for(int32 x = 0; x < width; x += 128)
   {
    const int32 chunk_x_max = std::min<int32>(width - x, 128);
-   uint16 tmpbuf[128];	// TODO: Check and see if the GPU is actually (ab)using the CLUT or texture cache.
+   uint16 tmpbuf[128];	// TODO: Check and see if the GPU is actually (ab)using the texture cache(doesn't seem to be affecting CLUT cache...).
 
    for(int32 chunk_x = 0; chunk_x < chunk_x_max; chunk_x++)
    {
@@ -566,7 +416,6 @@ INLINE void PS_GPU::Command_FBCopy(const uint32 *cb)
    }
   }
  }
-
 }
 
 INLINE void PS_GPU::Command_FBWrite(const uint32 *cb)
@@ -576,22 +425,28 @@ INLINE void PS_GPU::Command_FBWrite(const uint32 *cb)
  FBRW_X = (cb[1] >>  0) & 0x3FF;
  FBRW_Y = (cb[1] >> 16) & 0x3FF;
 
- FBRW_W = (cb[2] >>  0) & 0x7FF;
- FBRW_H = (cb[2] >> 16) & 0x3FF;
+ FBRW_W = (cb[2] >>  0) & 0x3FF;
+ FBRW_H = (cb[2] >> 16) & 0x1FF;
 
- if(FBRW_W > 0x400)
-  FBRW_W &= 0x3FF;
+ if(!FBRW_W)
+  FBRW_W = 0x400;
 
- if(FBRW_H > 0x200)
-  FBRW_H &= 0x1FF;
+ if(!FBRW_H)
+  FBRW_H = 0x200;
 
  FBRW_CurX = FBRW_X;
  FBRW_CurY = FBRW_Y;
 
+ InvalidateTexCache();
+
  if(FBRW_W != 0 && FBRW_H != 0)
   InCmd = INCMD_FBWRITE;
 }
 
+//
+// FBRead: PS1 GPU in SCPH-5501 gives odd, inconsistent results when raw_height == 0, or
+// raw_height != 0x200 && (raw_height & 0x1FF) == 0
+//
 INLINE void PS_GPU::Command_FBRead(const uint32 *cb)
 {
  assert(InCmd == INCMD_NONE);
@@ -599,11 +454,11 @@ INLINE void PS_GPU::Command_FBRead(const uint32 *cb)
  FBRW_X = (cb[1] >>  0) & 0x3FF;
  FBRW_Y = (cb[1] >> 16) & 0x3FF;
 
- FBRW_W = (cb[2] >>  0) & 0x7FF;
+ FBRW_W = (cb[2] >>  0) & 0x3FF;
  FBRW_H = (cb[2] >> 16) & 0x3FF;
 
- if(FBRW_W > 0x400)
-  FBRW_W &= 0x3FF;
+ if(!FBRW_W)
+  FBRW_W = 0x400;
 
  if(FBRW_H > 0x200)
   FBRW_H &= 0x1FF;
@@ -611,51 +466,63 @@ INLINE void PS_GPU::Command_FBRead(const uint32 *cb)
  FBRW_CurX = FBRW_X;
  FBRW_CurY = FBRW_Y;
 
+ InvalidateTexCache();
+
  if(FBRW_W != 0 && FBRW_H != 0)
   InCmd = INCMD_FBREAD;
 }
 
-
-INLINE void PS_GPU::RecalcTexWindowLUT(void)
+/*
+INLINE void PS_GPU::RecalcTexPageStuff(uint32 tpage)
 {
- const unsigned TexWindowX_AND = ~(tww << 3);
- const unsigned TexWindowX_OR = (twx & tww) << 3;
 
- const unsigned TexWindowY_AND = ~(twh << 3);
- const unsigned TexWindowY_OR = (twy & twh) << 3;
 
-// printf("TWX: 0x%02x, TWW: 0x%02x\n", twx, tww);
-// printf("TWY: 0x%02x, TWH: 0x%02x\n", twy, twh);
+}
+*/
 
- for(unsigned x = 0; x < 256; x++)
+INLINE void PS_GPU::SetTPage(const uint32 cmdw)
+{
+ const unsigned NewTexPageX = (cmdw & 0xF) * 64;
+ const unsigned NewTexPageY = (cmdw & 0x10) * 16;
+ const unsigned NewTexMode = (cmdw >> 7) & 0x3;
+
+ abr = (cmdw >> 5) & 0x3;
+
+ if(!NewTexMode != !TexMode || NewTexPageX != TexPageX || NewTexPageY != TexPageY)
  {
-  TexWindowXLUT[x] = (x & TexWindowX_AND) | TexWindowX_OR;
+  InvalidateTexCache();
  }
 
- for(unsigned y = 0; y < 256; y++)
+ if(TexDisableAllowChange)
  {
-  TexWindowYLUT[y] = (y & TexWindowY_AND) | TexWindowY_OR;
+  bool NewTexDisable = (cmdw >> 11) & 1;
+
+  if(NewTexDisable != TexDisable)
+   InvalidateTexCache();
+
+  TexDisable = NewTexDisable;
+  //printf("TexDisable: %02x\n", TexDisable);
  }
 
- memset(TexWindowXLUT_Pre, TexWindowXLUT[0], sizeof(TexWindowXLUT_Pre));
- memset(TexWindowXLUT_Post, TexWindowXLUT[255], sizeof(TexWindowXLUT_Post));
+ TexPageX = NewTexPageX;
+ TexPageY = NewTexPageY;
+ TexMode = NewTexMode;
 
- memset(TexWindowYLUT_Pre, TexWindowYLUT[0], sizeof(TexWindowYLUT_Pre));
- memset(TexWindowYLUT_Post, TexWindowYLUT[255], sizeof(TexWindowYLUT_Post));
+ //
+ //
+ RecalcTexWindowStuff();
 }
 
 INLINE void PS_GPU::Command_DrawMode(const uint32 *cb)
 {
- TexPageX = (*cb & 0xF) * 64;
- TexPageY = (*cb & 0x10) * 16;
+ const uint32 cmdw = *cb;
 
- SpriteFlip = *cb & 0x3000;
+ SetTPage(cmdw);
 
- abr = (*cb >> 5) & 0x3;
- TexMode = (*cb >> 7) & 0x3;
+ SpriteFlip = cmdw & 0x3000;
+ dtd = (cmdw >> 9) & 1;
+ dfe = (cmdw >> 10) & 1;
 
- dtd = (*cb >> 9) & 1;
- dfe = (*cb >> 10) & 1;
  //printf("*******************DFE: %d -- scanline=%d\n", dfe, scanline);
 }
 
@@ -666,19 +533,23 @@ INLINE void PS_GPU::Command_TexWindow(const uint32 *cb)
  twx = ((*cb >> 10) & 0x1F);
  twy = ((*cb >> 15) & 0x1F);
 
- RecalcTexWindowLUT();
+ RecalcTexWindowStuff();
 }
 
 INLINE void PS_GPU::Command_Clip0(const uint32 *cb)
 {
  ClipX0 = *cb & 1023;
  ClipY0 = (*cb >> 10) & 1023;
+
+ //fprintf(stderr, "[GPU] Clip0: x=%d y=%d, raw=0x%08x --- %d\n", ClipX0, ClipY0, *cb, scanline);
 }
 
 INLINE void PS_GPU::Command_Clip1(const uint32 *cb)
 {
  ClipX1 = *cb & 1023;
  ClipY1 = (*cb >> 10) & 1023;
+
+ //fprintf(stderr, "[GPU] Clip1: x=%d y=%d, raw=0x%08x --- %d\n", ClipX1, ClipY1, *cb, scanline);
 }
 
 INLINE void PS_GPU::Command_DrawingOffset(const uint32 *cb)
@@ -686,7 +557,7 @@ INLINE void PS_GPU::Command_DrawingOffset(const uint32 *cb)
  OffsX = sign_x_to_s32(11, (*cb & 2047));
  OffsY = sign_x_to_s32(11, ((*cb >> 11) & 2047));
 
- //fprintf(stderr, "[GPU] Drawing offset: %d(raw=%d) %d(raw=%d) -- %d\n", OffsX, *cb, OffsY, *cb >> 11, scanline);
+ //fprintf(stderr, "[GPU] Drawing offset: x=%d y=%d, raw=0x%08x --- %d\n", OffsX, OffsY, *cb, scanline);
 }
 
 INLINE void PS_GPU::Command_MaskSetting(const uint32 *cb)
@@ -696,9 +567,22 @@ INLINE void PS_GPU::Command_MaskSetting(const uint32 *cb)
  MaskEvalAND = (*cb & 2) ? 0x8000 : 0x0000;
 }
 
+INLINE void PS_GPU::InvalidateTexCache(void)
+{
+ for(auto& c : TexCache)
+  c.Tag = ~0U;
+}
+
+void PS_GPU::InvalidateCache(void)
+{
+ CLUT_Cache_VB = ~0U;
+
+ InvalidateTexCache();
+}
+
 INLINE void PS_GPU::Command_ClearCache(const uint32 *cb)
 {
-
+ InvalidateCache();
 }
 
 INLINE void PS_GPU::Command_IRQ(const uint32 *cb)
@@ -710,24 +594,6 @@ INLINE void PS_GPU::Command_IRQ(const uint32 *cb)
 //
 // C-style function wrappers so our command table isn't so ginormous(in memory usage).
 //
-template<int numvertices, bool shaded, bool textured, int BlendMode, bool TexMult, uint32 TexMode_TA, bool MaskEval_TA>
-static void G_Command_DrawPolygon(PS_GPU* g, const uint32 *cb)
-{
- g->Command_DrawPolygon<numvertices, shaded, textured, BlendMode, TexMult, TexMode_TA, MaskEval_TA>(cb);
-}
-
-template<uint8 raw_size, bool textured, int BlendMode, bool TexMult, uint32 TexMode_TA, bool MaskEval_TA>
-static void G_Command_DrawSprite(PS_GPU* g, const uint32 *cb)
-{
- g->Command_DrawSprite<raw_size, textured, BlendMode, TexMult, TexMode_TA, MaskEval_TA>(cb);
-}
-
-template<bool polyline, bool goraud, int BlendMode, bool MaskEval_TA>
-static void G_Command_DrawLine(PS_GPU* g, const uint32 *cb)
-{
- g->Command_DrawLine<polyline, goraud, BlendMode, MaskEval_TA>(cb);
-}
-
 static void G_Command_ClearCache(PS_GPU* g, const uint32 *cb)
 {
  g->Command_ClearCache(cb);
@@ -788,10 +654,53 @@ static void G_Command_MaskSetting(PS_GPU* g, const uint32 *cb)
  g->Command_MaskSetting(cb);
 }
 
+CTEntry PS_GPU::Commands[0x100];
 
-CTEntry PS_GPU::Commands[256] =
+const CTEntry PS_GPU::Commands_00_1F[0x20] =
 {
- #include "gpu_command_table.inc"
+ /* 0x00 */
+ NULLCMD(),
+ OTHER_HELPER(1, 2, false, G_Command_ClearCache),
+ OTHER_HELPER(3, 3, false, G_Command_FBFill),
+
+ NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(),
+ NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(),
+
+ /* 0x10 */
+ NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(),
+ NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(),
+
+ /* 0x1F */
+ OTHER_HELPER(1, 1, false,  G_Command_IRQ)
+};
+
+const CTEntry PS_GPU::Commands_80_FF[0x80] =
+{
+ /* 0x80 ... 0x9F */
+ OTHER_HELPER_X32(4, 2, false, G_Command_FBCopy),
+
+ /* 0xA0 ... 0xBF */
+ OTHER_HELPER_X32(3, 2, false, G_Command_FBWrite),
+
+ /* 0xC0 ... 0xDF */
+ OTHER_HELPER_X32(3, 2, false, G_Command_FBRead),
+
+ /* 0xE0 */
+
+ NULLCMD(),
+ OTHER_HELPER(1, 2, false, G_Command_DrawMode),
+ OTHER_HELPER(1, 2, false, G_Command_TexWindow),
+ OTHER_HELPER(1, 1, true,  G_Command_Clip0),
+ OTHER_HELPER(1, 1, true,  G_Command_Clip1),
+ OTHER_HELPER(1, 1, true,  G_Command_DrawingOffset),
+ OTHER_HELPER(1, 2, false, G_Command_MaskSetting),
+
+ NULLCMD(),
+ NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(),
+
+ /* 0xF0 */
+ NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(),
+ NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD(), NULLCMD()
 };
 
 void PS_GPU::ProcessFIFO(void)
@@ -814,7 +723,7 @@ void PS_GPU::ProcessFIFO(void)
 
   case INCMD_FBWRITE:
        {
-  	uint32 InData = BlitterFIFO.ReadUnit();
+  	uint32 InData = BlitterFIFO.Read();
 
   	for(int i = 0; i < 2; i++)
   	{
@@ -852,7 +761,7 @@ void PS_GPU::ProcessFIFO(void)
 	{
 	 for(unsigned i = 0; i < vl; i++)
 	 {
-	  CB[i] = BlitterFIFO.ReadUnit();
+	  CB[i] = BlitterFIFO.Read();
 	 }
 
 	 command->func[abr][TexMode | (MaskEvalAND ? 0x4 : 0x0)](this, CB);
@@ -871,9 +780,9 @@ void PS_GPU::ProcessFIFO(void)
 	unsigned vl = 1 + (bool)(InCmd_CC & 0x10);
 	uint32 CB[2];
 
-  	if((BlitterFIFO.ReadUnit(true) & 0xF000F000) == 0x50005000)
+  	if((BlitterFIFO.Peek() & 0xF000F000) == 0x50005000)
 	{
-	 BlitterFIFO.ReadUnit();
+	 BlitterFIFO.Read();
 	 InCmd = INCMD_NONE;
 	 return;
 	}
@@ -882,7 +791,7 @@ void PS_GPU::ProcessFIFO(void)
 	{
 	 for(unsigned i = 0; i < vl; i++)
 	 {
-	  CB[i] = BlitterFIFO.ReadUnit();
+	  CB[i] = BlitterFIFO.Read();
 	 }
 
 	 command->func[abr][TexMode | (MaskEvalAND ? 0x4 : 0x0)](this, CB);
@@ -892,7 +801,7 @@ void PS_GPU::ProcessFIFO(void)
        break;
  }
 
- const uint32 cc = BlitterFIFO.ReadUnit(true) >> 24;
+ const uint32 cc = BlitterFIFO.Peek() >> 24;
  const CTEntry *command = &Commands[cc];
 
  if(DrawTimeAvail < 0 && !command->ss_cmd)
@@ -903,7 +812,7 @@ void PS_GPU::ProcessFIFO(void)
   uint32 CB[0x10];
 
   for(unsigned i = 0; i < command->len; i++)
-   CB[i] = BlitterFIFO.ReadUnit();
+   CB[i] = BlitterFIFO.Read();
 
   if(!command->ss_cmd)
    DrawTimeAvail -= 2;
@@ -921,17 +830,10 @@ void PS_GPU::ProcessFIFO(void)
   // A very very ugly kludge to support texture mode specialization. fixme/cleanup/SOMETHING in the future.
   if(cc >= 0x20 && cc <= 0x3F && (cc & 0x4))
   {
-   uint32 tpage;
-
-   tpage = CB[4 + ((cc >> 4) & 0x1)] >> 16;
-
-   TexPageX = (tpage & 0xF) * 64;
-   TexPageY = (tpage & 0x10) * 16;
-
-   SpriteFlip = tpage & 0x3000;
-
-   abr = (tpage >> 5) & 0x3;
-   TexMode = (tpage >> 7) & 0x3;
+   //
+   // Don't alter SpriteFlip here.
+   //
+   SetTPage(CB[4 + ((cc >> 4) & 0x1)] >> 16);
   }
 
   if(!command->func[abr][TexMode])
@@ -948,13 +850,13 @@ void PS_GPU::ProcessFIFO(void)
 
 INLINE void PS_GPU::WriteCB(uint32 InData)
 {
- if(BlitterFIFO.CanRead() >= 0x10 && (InCmd != INCMD_NONE || (BlitterFIFO.CanRead() - 0x10) >= Commands[BlitterFIFO.ReadUnit(true) >> 24].fifo_fb_len))
+ if(BlitterFIFO.CanRead() >= 0x10 && (InCmd != INCMD_NONE || (BlitterFIFO.CanRead() - 0x10) >= Commands[BlitterFIFO.Peek() >> 24].fifo_fb_len))
  {
   PSX_DBG(PSX_DBG_WARNING, "GPU FIFO overflow!!!\n");
   return;
  }
 
- BlitterFIFO.WriteUnit(InData);
+ BlitterFIFO.Write(InData);
  ProcessFIFO();
 }
 
@@ -1023,30 +925,45 @@ void PS_GPU::Write(const pscpu_timestamp_t timestamp, uint32 A, uint32 V)
 	DisplayMode = V & 0xFF;
 	break;
 
+   case 0x09:
+	TexDisableAllowChange = V & 1;
+	break;
+
    case 0x10:	// GPU info(?)
 	switch(V & 0xF)
 	{
 	 // DataReadBuffer must remain unchanged for any unhandled GPU info index.
 	 default:  break;
 
-	 case 0x2: DataReadBuffer = (tww << 0) | (twh << 5) | (twx << 10) | (twy << 15);
+	 case 0x2: DataReadBufferEx &= 0xFFF00000;
+		   DataReadBufferEx |= (tww << 0) | (twh << 5) | (twx << 10) | (twy << 15);
+		   DataReadBuffer = DataReadBufferEx;
 		   break;
 
-	 case 0x3: DataReadBuffer = (ClipY0 << 10) | ClipX0;
+	 case 0x3: DataReadBufferEx &= 0xFFF00000;
+		   DataReadBufferEx |= (ClipY0 << 10) | ClipX0;
+		   DataReadBuffer = DataReadBufferEx;
 		   break;
 
- 	 case 0x4: DataReadBuffer = (ClipY1 << 10) | ClipX1;
+ 	 case 0x4: DataReadBufferEx &= 0xFFF00000;
+		   DataReadBufferEx |= (ClipY1 << 10) | ClipX1;
+		   DataReadBuffer = DataReadBufferEx;
 		   break;
 
-	 case 0x5: DataReadBuffer = (OffsX & 2047) | ((OffsY & 2047) << 11);
+	 case 0x5: DataReadBufferEx &= 0xFFC00000;
+		   DataReadBufferEx |= (OffsX & 2047) | ((OffsY & 2047) << 11);
+		   DataReadBuffer = DataReadBufferEx;
 		   break;
 
-	 case 0x7: DataReadBuffer = 2;
+	 case 0x7: DataReadBufferEx = 2;
+		   DataReadBuffer = DataReadBufferEx;
 		   break;
 
-	 case 0x8: DataReadBuffer = 0;
+	 case 0x8: DataReadBufferEx = 0;
+		   DataReadBuffer = DataReadBufferEx;
 		   break;
 	}
+	//fprintf(stderr, "[GPU] CC 0x10:0x%02x, DRB=0x%02x\n", V & 0xF, DataReadBuffer);
 	break;
 
   }
@@ -1070,23 +987,27 @@ INLINE uint32 PS_GPU::ReadData(void)
 {
  if(InCmd == INCMD_FBREAD)
  {
-  DataReadBuffer = 0;
+  DataReadBufferEx = 0;
   for(int i = 0; i < 2; i++)
   {
-   DataReadBuffer |= GPURAM[FBRW_CurY & 511][FBRW_CurX & 1023] << (i * 16);
+   DataReadBufferEx |= GPURAM[FBRW_CurY & 511][FBRW_CurX & 1023] << (i * 16);
 
    FBRW_CurX++;
    if(FBRW_CurX == (FBRW_X + FBRW_W))
    {
-    FBRW_CurX = FBRW_X;
-    FBRW_CurY++;
-    if(FBRW_CurY == (FBRW_Y + FBRW_H))
+    if((FBRW_CurY + 1) == (FBRW_Y + FBRW_H))
     {
      InCmd = INCMD_NONE;
-     break;
+    }
+    else
+    {
+     FBRW_CurY++;
+     FBRW_CurX = FBRW_X; 
     }
    }
   }
+
+  return DataReadBufferEx;
  }
 
  return DataReadBuffer;
@@ -1105,6 +1026,8 @@ uint32 PS_GPU::Read(const pscpu_timestamp_t timestamp, uint32 A)
  {
   ret = (((DisplayMode << 1) & 0x7F) | ((DisplayMode >> 6) & 1)) << 16;
 
+  ret |= (DisplayMode & 0x80) << 7;
+
   ret |= DMAControl << 29;
 
   ret |= (DisplayFB_CurLineYReadout & 1) << 31;
@@ -1141,6 +1064,8 @@ uint32 PS_GPU::Read(const pscpu_timestamp_t timestamp, uint32 A)
 
   if(MaskEvalAND)
    ret |= 1 << 12;
+
+  ret |= TexDisable << 15;
  }
  else		// "Data"
   ret = ReadData();
@@ -1153,20 +1078,23 @@ uint32 PS_GPU::Read(const pscpu_timestamp_t timestamp, uint32 A)
  return(ret >> ((A & 3) * 8));
 }
 
-/*
-static INLINE uint32 ShiftHelper(uint32 val, int shamt, uint32 mask)
+#if 0
+static INLINE uint32 MDFN_NOWARN_UNUSED ShiftHelper(uint32 val, int shamt, uint32 mask)
 {
  if(shamt < 0)
   return((val >> (-shamt)) & mask);
  else
   return((val << shamt) & mask);
 }
-*/
+#endif
+
+#pragma GCC push_options
+#pragma GCC optimize("no-unroll-loops,no-peel-loops,no-crossjumping")
 INLINE void PS_GPU::ReorderRGB_Var(uint32 out_Rshift, uint32 out_Gshift, uint32 out_Bshift, bool bpp24, const uint16 *src, uint32 *dest, const int32 dx_start, const int32 dx_end, int32 fb_x)
 {
      if(bpp24)	// 24bpp
      {
-      for(int32 x = dx_start; x < dx_end; x++)
+      for(int32 x = dx_start; MDFN_LIKELY(x < dx_end); x++)
       {
        uint32 srcpix;
 
@@ -1181,33 +1109,37 @@ INLINE void PS_GPU::ReorderRGB_Var(uint32 out_Rshift, uint32 out_Gshift, uint32
      }				// 15bpp
      else
      {
-      for(int32 x = dx_start; x < dx_end; x++)
+      for(int32 x = dx_start; MDFN_LIKELY(x < dx_end); x++)
       {
        uint32 srcpix = src[fb_x >> 1];
 
-       dest[x] = OutputLUT[srcpix & 0x7FFF];
-       //dest[x] = ShiftHelper(srcpix, out_Rshift + 3 -  0, (0xF8 << out_Rshift)) |
-//	         ShiftHelper(srcpix, out_Gshift + 3 -  5, (0xF8 << out_Gshift)) |
-//	         ShiftHelper(srcpix, out_Bshift + 3 - 10, (0xF8 << out_Bshift));
-
+#if 1
+       dest[x] = OutputLUT[(uint8)srcpix] | (OutputLUT + 256)[(srcpix >> 8) & 0x7F];
+#else
+       dest[x] = ShiftHelper(srcpix, out_Rshift + 3 -  0, (0xF8 << out_Rshift)) |
+	         ShiftHelper(srcpix, out_Gshift + 3 -  5, (0xF8 << out_Gshift)) |
+	         ShiftHelper(srcpix, out_Bshift + 3 - 10, (0xF8 << out_Bshift));
+#endif
        fb_x = (fb_x + 2) & 0x7FF;
       }
      }
 
 }
 
-
 template<uint32 out_Rshift, uint32 out_Gshift, uint32 out_Bshift>
-void PS_GPU::ReorderRGB(bool bpp24, const uint16 *src, uint32 *dest, const int32 dx_start, const int32 dx_end, int32 fb_x)
+void NO_INLINE PS_GPU::ReorderRGB(bool bpp24, const uint16 *src, uint32 *dest, const int32 dx_start, const int32 dx_end, int32 fb_x)
 {
  ReorderRGB_Var(out_Rshift, out_Gshift, out_Bshift, bpp24, src, dest, dx_start, dx_end, fb_x);
 }
+#pragma GCC pop_options
 
 pscpu_timestamp_t PS_GPU::Update(const pscpu_timestamp_t sys_timestamp)
 {
  static const uint32 DotClockRatios[5] = { 10, 8, 5, 4, 7 };
  const uint32 dmc = (DisplayMode & 0x40) ? 4 : (DisplayMode & 0x3);
- const uint32 dmw = 2800 / DotClockRatios[dmc];	// Must be <= 768
+ const uint32 dmw = 2800 / DotClockRatios[dmc];	// Must be <= (768 - 32)
+ const uint32 dmpa = (2800 - (hide_hoverscan ? 2640 : 2800)) / DotClockRatios[dmc] / 2;	// Must be <= 32
+ const uint32 drxbo = 32;
 
  int32 sys_clocks = sys_timestamp - lastts;
  int32 gpu_clocks;
@@ -1355,11 +1287,14 @@ pscpu_timestamp_t PS_GPU::Update(const pscpu_timestamp_t sys_timestamp)
          dest[x] = black;
         }
        }
-       char buffer[256];
 
-       trio_snprintf(buffer, sizeof(buffer), _("VIDEO STANDARD MISMATCH"));
-       DrawTextTrans(surface->pixels + ((DisplayRect->h / 2) - (13 / 2)) * surface->pitch32, surface->pitch32 << 2, DisplayRect->w, (UTF8*)buffer,
+       if(!DisplayOff)
+       {
+        char buffer[256];
+        trio_snprintf(buffer, sizeof(buffer), _("VIDEO STANDARD MISMATCH"));
+        DrawTextTrans(surface->pixels + ((DisplayRect->h / 2) - (13 / 2)) * surface->pitch32, surface->pitch32 << 2, DisplayRect->w, buffer,
 		surface->MakeColor(0x00, 0xFF, 0x00), true, MDFN_FONT_6x13_12x13);
+       }
       }
       else
       {
@@ -1368,7 +1303,7 @@ pscpu_timestamp_t PS_GPU::Update(const pscpu_timestamp_t sys_timestamp)
        espec->InterlaceOn = (bool)(DisplayMode & 0x20);
        espec->InterlaceField = (bool)(DisplayMode & 0x20) && field;
 
-       DisplayRect->x = 0;
+       DisplayRect->x = drxbo;
        DisplayRect->y = 0;
        DisplayRect->w = 0;
        DisplayRect->h = VisibleLineCount << (bool)(DisplayMode & 0x20);
@@ -1395,14 +1330,18 @@ pscpu_timestamp_t PS_GPU::Update(const pscpu_timestamp_t sys_timestamp)
      if(sl_zero_reached)
      {
       // Gameplay in Descent(NTSC) has vblank at scanline 236
-      if(scanline >= (FirstVisibleLine + VisibleLineCount) || (scanline >= (HardwarePALType ? 260 : 232)))
+      //
+      // Mikagura Shoujo Tanteidan has vblank at scanline 192 during intro
+      //  FMV(which we don't handle here because low-latency in that case is not so important).
+      //
+      if(scanline >= (HardwarePALType ? 260 : 232))
       {
        //printf("Req Exit(vblank case): %u\n", scanline);
        PSX_RequestMLExit();
       }
       else
       {
-       //printf("VBlank too early, chickening out early exit!\n");
+       //printf("VBlank too early, chickening out early exit: %u!\n", scanline);
       }
      }
 
@@ -1442,13 +1381,13 @@ pscpu_timestamp_t PS_GPU::Update(const pscpu_timestamp_t sys_timestamp)
 
     if((bool)(DisplayMode & 0x08) == HardwarePALType && scanline >= FirstVisibleLine && scanline < (FirstVisibleLine + VisibleLineCount) && !skip && espec)
     {
-     uint32 *dest;	// = surface->pixels + (scanline - VisibleStartLine) * surface->pitch32;
+     uint32 *dest;
      int32 dest_line;
      int32 fb_x = DisplayFB_XStart * 2;
      int32 dx_start = HorizStart, dx_end = HorizEnd;
 
      dest_line = ((scanline - FirstVisibleLine) << espec->InterlaceOn) + espec->InterlaceField;
-     dest = surface->pixels + dest_line * surface->pitch32;
+     dest = surface->pixels + (drxbo - dmpa) + dest_line * surface->pitch32;
 
      if(dx_end < dx_start)
       dx_end = dx_start;
@@ -1456,8 +1395,10 @@ pscpu_timestamp_t PS_GPU::Update(const pscpu_timestamp_t sys_timestamp)
      dx_start = dx_start / DotClockRatios[dmc];
      dx_end = dx_end / DotClockRatios[dmc];
 
-     dx_start -= 488 / DotClockRatios[dmc];
-     dx_end -= 488 / DotClockRatios[dmc];
+     dx_start -= hmc_to_visible / DotClockRatios[dmc];
+     dx_end -= hmc_to_visible / DotClockRatios[dmc];
+     dx_start += 7;
+     dx_end += 7;
 
      if(dx_start < 0)
      {
@@ -1472,11 +1413,7 @@ pscpu_timestamp_t PS_GPU::Update(const pscpu_timestamp_t sys_timestamp)
      if(InVBlank || DisplayOff)
       dx_start = dx_end = 0;
 
-     // TODO, but there are problems with this, as not all blitter busy cycles(crudely abstracted with DrawTimeAvail) are GPU RAM access cycles.
-     // Also, it shouldn't be here per-se, since this code won't be all if we're frameskipping or there's a video standard mismatch
-     //DrawTimeAvail -= (dx_end - dx_start) + ((DisplayMode & 0x10) ? ((dx_end - dx_start + 1) >> 1) : 0);
-
-     LineWidths[dest_line] = dmw;
+     LineWidths[dest_line] = dmw - dmpa * 2;
 
      {
       const uint16 *src = GPURAM[DisplayFB_CurLineYReadout];
@@ -1504,7 +1441,7 @@ pscpu_timestamp_t PS_GPU::Update(const pscpu_timestamp_t sys_timestamp)
      //if(scanline == 64)
      // printf("%u\n", sys_timestamp - ((uint64)gpu_clocks * 65536) / GPUClockRatio);
 
-     PSX_GPULineHook(sys_timestamp, sys_timestamp - ((uint64)gpu_clocks * 65536) / GPUClockRatio, scanline == 0, dest, &surface->format, dmw, (488 - 146) / DotClockRatios[dmc], (HardwarePALType ? 53203425 : 53693182) / DotClockRatios[dmc], DotClockRatios[dmc]);
+     PSX_GPULineHook(sys_timestamp, sys_timestamp - ((uint64)gpu_clocks * 65536) / GPUClockRatio, scanline == 0, dest, &surface->format, dmw, (hmc_to_visible - 220) / DotClockRatios[dmc], (HardwarePALType ? 53203425 : 53693182) / DotClockRatios[dmc], DotClockRatios[dmc]);
     }
     else
     {
@@ -1516,8 +1453,10 @@ pscpu_timestamp_t PS_GPU::Update(const pscpu_timestamp_t sys_timestamp)
      DisplayFB_CurYOffset = (DisplayFB_CurYOffset + 1) & 0x1FF;
     }
    }
-  }
- }
+   PSX_SetEventNT(PSX_EVENT_TIMER, TIMER_Update(sys_timestamp));  // Mostly so the next event time gets recalculated properly in regards to our calls
+								  // to TIMER_SetVBlank() and TIMER_SetHRetrace().
+  }	// end if(!LineClockCounter)
+ }	// end while(gpu_clocks > 0)
 
  //puts("GPU Update End");
 
@@ -1561,24 +1500,42 @@ void PS_GPU::StartFrame(EmulateSpecStruct *espec_arg)
 
  if(espec->VideoFormatChanged)
  {
+  const auto& f = surface->format;
+
   for(int rc = 0; rc < 0x8000; rc++)
   {
-   uint32 r, g, b;
+   const uint8 a = rc;
+   const uint8 b = rc >> 8;
 
-   r = ((rc >> 0) & 0x1F) << 3;
-   g = ((rc >> 5) & 0x1F) << 3;
-   b = ((rc >> 10) & 0x1F) << 3;
-   OutputLUT[rc] = espec->surface->format.MakeColor(r, g, b, 0);
+   (OutputLUT +   0)[a] = ((a & 0x1F) << (3 + f.Rshift)) | ((a >> 5) << (3 + f.Gshift));
+   (OutputLUT + 256)[b] = ((b & 0x3) << (6 + f.Gshift)) | (((b >> 2) & 0x1F) << (3 + f.Bshift));
   }
  }
 }
 
-int PS_GPU::StateAction(StateMem *sm, int load, int data_only)
+void PS_GPU::StateAction(StateMem *sm, const unsigned load, const bool data_only)
 {
+ uint32 TexCache_Tag[256];
+ uint16 TexCache_Data[256][4];
+
+ for(unsigned i = 0; i < 256; i++)
+ {
+  TexCache_Tag[i] = TexCache[i].Tag;
+
+  for(unsigned j = 0; j < 4; j++)
+   TexCache_Data[i][j] = TexCache[i].Data[j];
+ }
+
  SFORMAT StateRegs[] =
  {
   SFARRAY16(&GPURAM[0][0], sizeof(GPURAM) / sizeof(GPURAM[0][0])),
 
+  SFARRAY16(&CLUT_Cache[0], sizeof(CLUT_Cache) / sizeof(CLUT_Cache[0])),
+  SFVAR(CLUT_Cache_VB),
+
+  SFARRAY32(TexCache_Tag, sizeof(TexCache_Tag) / sizeof(TexCache_Tag[0])),
+  SFARRAY16(&TexCache_Data[0][0], sizeof(TexCache_Data) / sizeof(TexCache_Data[0][0])),
+
   SFVAR(DMAControl),
 
   SFVAR(ClipX0),
@@ -1595,6 +1552,9 @@ int PS_GPU::StateAction(StateMem *sm, int load, int data_only)
   SFVAR(MaskSetOR),
   SFVAR(MaskEvalAND),
 
+  SFVAR(TexDisable),
+  SFVAR(TexDisableAllowChange),
+
   SFVAR(tww),
   SFVAR(twh),
   SFVAR(twx),
@@ -1608,12 +1568,13 @@ int PS_GPU::StateAction(StateMem *sm, int load, int data_only)
   SFVAR(abr),
   SFVAR(TexMode),
 
-  SFARRAY32(&BlitterFIFO.data[0], BlitterFIFO.data.size()),
+  SFARRAY32(&BlitterFIFO.data[0], sizeof(BlitterFIFO.data) / sizeof(BlitterFIFO.data[0])),
   SFVAR(BlitterFIFO.read_pos),
   SFVAR(BlitterFIFO.write_pos),
   SFVAR(BlitterFIFO.in_count),
 
   SFVAR(DataReadBuffer),
+  SFVAR(DataReadBufferEx),
 
   SFVAR(IRQPending),
 
@@ -1625,7 +1586,6 @@ int PS_GPU::StateAction(StateMem *sm, int load, int data_only)
   TVHELPER(InQuad_F3Vertices[1]),
   TVHELPER(InQuad_F3Vertices[2]),
 #undef TVHELPER
-  SFVAR(InQuad_clut),
 
   SFVAR(InPLine_PrevPoint.x),
   SFVAR(InPLine_PrevPoint.y),
@@ -1672,11 +1632,20 @@ int PS_GPU::StateAction(StateMem *sm, int load, int data_only)
 
   SFEND
  };
- int ret = MDFNSS_StateAction(sm, load, data_only, StateRegs, "GPU");
+
+ MDFNSS_StateAction(sm, load, data_only, StateRegs, "GPU");
 
  if(load)
  {
-  RecalcTexWindowLUT();
+  for(unsigned i = 0; i < 256; i++)
+  {
+   TexCache[i].Tag = TexCache_Tag[i];
+
+   for(unsigned j = 0; j < 4; j++)
+    TexCache[i].Data[j] = TexCache_Data[i][j];
+  }
+
+  RecalcTexWindowStuff();
   BlitterFIFO.SaveStatePostLoad();
 
   HorizStart &= 0xFFF;
@@ -1684,8 +1653,6 @@ int PS_GPU::StateAction(StateMem *sm, int load, int data_only)
 
   IRQ_Assert(IRQ_GPU, IRQPending);
  }
-
- return(ret);
 }
 
 }
